@@ -13,15 +13,11 @@ import           Control.Monad (void)
 import           Control.Monad.Trans.Cont
 import           Data.ByteString (ByteString)
 import qualified Data.Map.Strict as Map
-import           Data.Maybe
-import           Data.Monoid
-import           Data.String
 import qualified Database.PostgreSQL.Simple as PG
 import qualified Database.PostgreSQL.Simple.Options as Client
-import           System.Environment
 import           System.Exit (ExitCode(..))
+import           System.IO.Unsafe (unsafePerformIO)
 import           System.Process
-import           System.Random
 import           Text.PrettyPrint.ANSI.Leijen hiding ((<$>))
 
 -- | Handle for holding temporary resources, the @postgres@ process handle
@@ -187,7 +183,6 @@ defaultConfig = mempty
   { plan = mempty
     { logger = pure mempty
     , postgresConfigFile = defaultPostgresConfig
-    , createDbConfig = Nothing
     , initDbConfig = pure mempty
       { commandLine = mempty
         { keyBased = Map.singleton "--no-sync" Nothing
@@ -320,7 +315,7 @@ startConfig extra = try $ evalContT $ do
   dbResources@Resources {..} <-
     ContT $ bracketOnError (setupConfig extra) cleanupConfig
   dbPostgresProcess <-
-    ContT $ bracketOnError (startPlan resourcesPlan) stopPostgresProcess
+    ContT $ bracketOnError (startPlan resourcesPlan) stopPlan
   pure DB {..}
 
 -- | Default start behavior. Equivalent to calling 'startConfig' with the
@@ -336,7 +331,7 @@ start = startConfig defaultConfig
 --   @since 1.12.0.0
 stop :: DB -> IO ()
 stop DB {..} = do
-  void $ stopPostgresProcess dbPostgresProcess
+  void $ stopPlan dbPostgresProcess
   cleanupConfig dbResources
 
 -- | Only stop the @postgres@ process but leave any temporary resources.
@@ -345,7 +340,15 @@ stop DB {..} = do
 --
 --   @since 1.12.0.0
 stopPostgres :: DB -> IO ExitCode
-stopPostgres = stopPostgresProcess . dbPostgresProcess
+stopPostgres = stopPlan . dbPostgresProcess
+
+-- | Only stop the @postgres@ process but leave any temporary resources.
+--   In contrast to 'stopPostgres' this function makes sure @postgres@
+--   has time to properly write files to the data directory.
+--
+--   @since 1.16.1.0
+stopPostgresGracefully :: DB -> IO ExitCode
+stopPostgresGracefully = stopPostgresProcess True . dbPostgresProcess
 
 -- | Restart the @postgres@ from 'DB' using the prior 'Plan'.
 --
@@ -356,7 +359,7 @@ restart db@DB{..} = try $ do
   let CompletePlan{..} = resourcesPlan dbResources
       startAction = startPostgresProcess completePlanConnectionTimeout completePlanLogger
         completePlanPostgres
-  bracketOnError startAction stopPostgresProcess $ \result ->
+  bracketOnError startAction stopPlan $ \result ->
     pure $ db { dbPostgresProcess = result }
 
 -- | Reload the configuration file without shutting down. Calls
@@ -433,190 +436,6 @@ optionsToDefaultConfig opts@Client.Options {..} =
 prettyPrintDB :: DB -> String
 prettyPrintDB = show . pretty
 
--------------------------------------------------------------------------------
--- withNewDb
--------------------------------------------------------------------------------
--- | Drop the db if it exists. Terminates all connections to the db first.
---
---   @since 1.12.0.0
-dropDbIfExists :: Client.Options -> String -> IO ()
-dropDbIfExists options dbName = do
-  let theConnectionString = Client.toConnectionString options
-      dropDbQuery = fromString $ "DROP DATABASE IF EXISTS " <> dbName <> ";"
-
-  terminateConnections $ options
-    { Client.dbname = pure dbName
-    }
-
-  mapException DeleteDbError $
-    bracket (PG.connectPostgreSQL theConnectionString) PG.close $
-      \conn -> void $ PG.execute_ conn dropDbQuery
-
-{-|
-Use the current database as a template and make a copy. Give the
-copy a random name.
-
-Equivalent to:
-
-@
- 'withNewDb' = 'withNewDbConfig' mempty
-@
-
-See 'startNewDbConfig' for more details.
-
-@since 1.12.0.0
--}
-withNewDb
-  :: DB
-  -- ^ The original 'DB' handle. The database name specified in the
-  --   connection options
-  --   is used as the template for the @generated@ 'ProcessConfig'
-  -> (DB -> IO a)
-  -- ^ The modified 'DB' handle that has the new database name
-  --   in it's connection options.
-  -> IO (Either StartError a)
-withNewDb = withNewDbConfig mempty
-
-{-|
-Exception safe version of 'startNewDbConfig'. Creates a
-temporary database using the current database as a template.
-
-See 'startNewDbConfig' for more details.
-
-@since 1.12.0.0
--}
-withNewDbConfig
-  :: ProcessConfig
-  -- ^ @extra@ @createdb@ 'ProcessConfig'.
-  -> DB
-  -- ^ The original 'DB' handle. The database name specified in the
-  --   connection options
-  --   is used as the template for the @generated@ 'ProcessConfig'.
-  -> (DB -> IO a)
-  -- ^ The modified 'DB' handle that has the new database name
-  --   in it's connection options.
-  -> IO (Either StartError a)
-withNewDbConfig extra db f =
-  bracket (startNewDbConfig extra db) (either mempty stopNewDb) $
-    either (pure . Left) (fmap Right . f)
-
-{-|
-Use the current database as a template and make a copy. Give the
-copy a random name.
-
-Equivalent to:
-
-@
- 'startNewDb' = 'startNewDbConfig' mempty
-@
-
-See 'startNewDbConfig' for more details.
-
-@since 1.13.0.0
--}
-startNewDb
-  :: DB
-  -- ^ The original 'DB' handle. The database name specified in the
-  --   connection options
-  --   is used as the template for the @generated@ 'ProcessConfig'.
-  -> IO (Either StartError DB)
-startNewDb = startNewDbConfig mempty
-
-{-|
-Use the current database as a template and make a copy. Give the
-copy a random name.
-
-Copying a database from a template can be faster than creating a new
-@postgres@ and migrating a database from scratch. In artifical benchmarks
-it appears to be about 2x faster.
-
-To use the current database as a template all connections to the database
-must be terminated first.
-
-To override the arguments passed to @createdb@ one can pass in @extra@
-'ProcessConfig'. The @combined@ process is created by 'mappend'ed the
-@generated@ with the @extra@ 'ProcessConfig', e.g.
-
-@
-   combined = generated '<>' extra
-@
-
-The current implementation has a few known issues.
-
-If a connection is made between the termination command and the @createdb@
-call the @createdb@ call will fail.
-
-Additionally the generated name is 32 character random name of characters
-\"a\" to \"z\". It is possible, although unlikeily that a duplicate
-database name could be generated and this would also cause a failure.
-
-'startNewDbConfig' requires cleanup so it best to use a 'bracket' along
-with 'stopNewDb'. This is equivalient to 'withNewDbConfig'.
-
-The only reason to use 'startNewDbConfig' over 'withNewDbConfig' is
-if you are unable to use 'withNewDbConfig' for some reason.
-
-@since 1.13.0.0
--}
-startNewDbConfig
-  :: ProcessConfig
-  -- ^ @extra@ @createdb@ 'ProcessConfig'.
-  -> DB
-  -- ^ The original 'DB' handle. The database name specified in the
-  --   connection options
-  --   is used as the template for the @generated@ 'ProcessConfig'.
-  -> IO (Either StartError DB)
-startNewDbConfig extra@ProcessConfig{..} db = try $ do
-  stdGen <- getStdGen
-  let oldOptions@Client.Options {..} = toConnectionOptions db
-      theDbName = fromMaybe "postgres" $ getLast dbname
-      newDbName = fromMaybe (take 32 $ randomRs ('a', 'z') stdGen) $
-        Map.lookup 0 $ indexBased commandLine
-      newOptions = oldOptions
-        { Client.dbname = pure newDbName
-        }
-      generated = standardProcessConfig
-        { commandLine = mempty
-          { keyBased = Map.fromList
-              [ ("-T", Just theDbName)
-              , ("-h", Just $ fromMaybe "127.0.0.1" $ getLast host)
-              , ("-p ", Just $ maybe "5432" show $ getLast port)
-              ]
-          , indexBased = Map.singleton 0 newDbName
-          }
-        }
-      combined = generated <> extra
-      newDb = db
-        { dbPostgresProcess = (dbPostgresProcess db)
-            { postgresProcessClientOptions = newOptions
-            }
-        }
-  envs <- getEnvironment
-  final <- case completeProcessConfig envs combined of
-    -- Failure case does not look possible
-    Left errs -> throwIO $ CompleteProcessConfigFailed (show $ pretty combined) errs
-    Right x -> pure x
-  terminateConnections oldOptions
-  executeCreateDb final
-  pure newDb
-
--- | Cleanup the temporary database created by 'startNewDbConfig'
---   or 'startNewDb'.
---
---   @since 1.13.0.0
-stopNewDb :: DB -> IO ()
-stopNewDb db = do
-  let oldOptions@Client.Options{..} = toConnectionOptions db
-  -- This should not happen.
-  newDbName <- maybe
-    (throwIO $ userError "stopNewDb: missing db name!")
-    pure $
-    getLast dbname
-  let template1Options = oldOptions
-        { Client.dbname = pure "template1"
-        }
-  dropDbIfExists template1Options newDbName
-
 -------------------------------------------------------------------------------
 -- initdb cache
 -------------------------------------------------------------------------------
@@ -640,19 +459,8 @@ data CacheConfig = CacheConfig
   --   and sets this to 'True' if it does.
   }
 
-{-|
-'createDefaultCacheConfig' attempts to determine if the @cp@ on the path
-supports \"copy on write\" flags and if it does, sets 'cacheUseCopyOnWrite'
-to 'True'.
-
-It sets 'cacheDirectoryType' to 'Permanent' @~\/.tmp-postgres@ and
-'cacheTemporaryDirectory' to @\/tmp@ (but this is not used when
-'Permanent' is set).
-
-@since 1.15.1.0
--}
-createDefaultCacheConfig :: IO CacheConfig
-createDefaultCacheConfig = do
+cowCheck :: Bool
+cowCheck = unsafePerformIO $ do
   let
 #ifdef darwin_HOST_OS
     cpFlag = "-c"
@@ -663,11 +471,28 @@ createDefaultCacheConfig = do
   -- if the flags do not exist we get a message like "cp: illegal option"
   let usage = "usage:" -- macos
       missingFile = "cp: missing file operand" -- linux
-      cacheUseCopyOnWrite = usage ==  take (length usage) errorOutput
-        || missingFile ==  take (length missingFile) errorOutput
-      cacheDirectoryType = Permanent "~/.tmp-postgres"
-      cacheTemporaryDirectory = "/tmp"
-  pure CacheConfig {..}
+  pure $ usage ==  take (length usage) errorOutput
+       || missingFile ==  take (length missingFile) errorOutput
+{-# NOINLINE cowCheck #-}
+
+{-|
+'defaultCacheConfig' attempts to determine if the @cp@ on the path
+supports \"copy on write\" flags and if it does, sets 'cacheUseCopyOnWrite'
+to 'True'.
+
+It sets 'cacheDirectoryType' to 'Permanent' @~\/.tmp-postgres@ and
+'cacheTemporaryDirectory' to @\/tmp@ (but this is not used when
+'Permanent' is set).
+
+@since 1.19.0.0
+-}
+defaultCacheConfig :: CacheConfig
+defaultCacheConfig = CacheConfig
+  { cacheDirectoryType = Permanent "~/.tmp-postgres"
+  , cacheTemporaryDirectory = "/tmp"
+  , cacheUseCopyOnWrite = cowCheck
+  }
+
 
 -- | Setup the @initdb@ cache folder.
 setupInitDbCache
@@ -712,13 +537,12 @@ withDbCacheConfig config =
 
 {-|
 Equivalent to 'withDbCacheConfig' with the 'CacheConfig'
-'createDefaultCacheConfig' makes.
+'defaultCacheConfig' makes.
 
 @since 1.15.1.0
 -}
 withDbCache :: ((Bool, CompleteDirectoryType) -> IO a) -> IO a
-withDbCache action =
-  flip withDbCacheConfig action =<< createDefaultCacheConfig
+withDbCache = withDbCacheConfig defaultCacheConfig
 
 {-|
 Helper to make a 'Config' out of caching info.
@@ -729,3 +553,84 @@ toCacheConfig :: (Bool, CompleteDirectoryType) -> Config
 toCacheConfig cacheInfo = mempty
   { initDbCache = pure $ pure $ fmap toFilePath cacheInfo
   }
+
+-------------------------------------------------------------------------------
+-- withSnapshot
+-------------------------------------------------------------------------------
+{- |
+Shutdown the database and copy the directory to a folder.
+
+@since 1.17.0.0
+-}
+takeSnapshot
+  :: DirectoryType
+  -- ^ Either a 'Temporary' or preexisting 'Permanent' directory.
+  -> DB
+  -- ^ The handle. The @postgres@ is shutdown and the data directory is copied.
+  -> IO (Either StartError CompleteDirectoryType)
+takeSnapshot directoryType db = try $ do
+  throwIfNotSuccess id =<< stopPostgresGracefully db
+  let
+#ifdef darwin_HOST_OS
+    cpFlags = if cowCheck then "cp -Rc " else "cp -R "
+#else
+    cpFlags = if cowCheck then "cp -R --reflink=auto " else "cp -R "
+#endif
+  bracketOnError
+    (setupDirectoryType
+      (toTemporaryDirectory db)
+      "tmp-postgres-snapshot"
+      directoryType
+    )
+    cleanupDirectoryType $ \snapShotDir -> do
+      let snapshotCopyCmd = cpFlags <>
+            toDataDirectory db <> "/* " <> toFilePath snapShotDir
+      throwIfNotSuccess (SnapshotCopyFailed snapshotCopyCmd) =<<
+        system snapshotCopyCmd
+
+      pure snapShotDir
+
+{-|
+Cleanup any temporary resources used for the snapshot.
+
+@since 1.17.0.0
+-}
+cleanupSnapshot :: CompleteDirectoryType -> IO ()
+cleanupSnapshot = cleanupDirectoryType
+
+{- |
+Exception safe method for taking a file system level copy of the database cluster.
+
+Snapshots are useful if you would like to start every test from a migrated database
+and the migration process is more time consuming then copying the additional data.
+
+@since 1.17.0.0
+-}
+withSnapshot
+  :: DirectoryType
+  -> DB
+  -> (CompleteDirectoryType -> IO a)
+  -> IO (Either StartError a)
+withSnapshot dirType db f = bracket
+  (takeSnapshot dirType db)
+  (either mempty cleanupSnapshot)
+  (either (pure . Left) (fmap Right . f))
+
+{-|
+Convert a snapshot into a 'Config' that includes a 'copyConfig' for copying the
+snapshot directory to a temporary directory.
+
+@since 1.17.0.0
+-}
+configFromSavePoint :: FilePath -> IO Config
+configFromSavePoint savePointPath = do
+  pure mempty
+    { plan = mempty
+        { copyConfig = pure $ pure CopyDirectoryCommand
+            { sourceDirectory = savePointPath
+            , destinationDirectory = Nothing
+            , useCopyOnWrite = cowCheck
+            }
+        , initDbConfig = Zlich
+        }
+    }
