@@ -21,7 +21,7 @@ import qualified Database.PostgreSQL.Simple.Options as Client
 import           System.Directory
 import           System.Exit (ExitCode(..))
 import           System.IO
-import           System.Posix.Signals (sigQUIT, signalProcess)
+import           System.Posix.Signals (sigINT, sigQUIT, signalProcess)
 import           System.Process
 import           System.Process.Internals
 import           System.Timeout
@@ -86,6 +86,8 @@ data StartError
   | FailedToFindDataDirectory String
   -- ^ Failed to find a data directory when trying to get
   --   a cached @initdb@ folder.
+  | SnapshotCopyFailed String ExitCode
+  -- ^ We tried to copy a data directory to a snapshot folder and it failed
   deriving (Show, Eq, Typeable)
 
 instance Exception StartError
@@ -251,12 +253,12 @@ terminateConnections options = do
 
 -- | Stop the @postgres@ process after attempting to terminate all the
 --   connections.
-stopPostgresProcess :: PostgresProcess -> IO ExitCode
-stopPostgresProcess PostgresProcess{..} = do
+stopPostgresProcess :: Bool -> PostgresProcess -> IO ExitCode
+stopPostgresProcess graceful PostgresProcess{..} = do
   withProcessHandle postgresProcessHandle $ \case
     OpenHandle p   ->
       -- Call for "Immediate shutdown"
-      signalProcess sigQUIT p
+      signalProcess (if graceful then sigINT else sigQUIT) p
     OpenExtHandle {} -> pure () -- TODO log windows is not supported
     ClosedHandle _ -> return ()
 
@@ -273,7 +275,7 @@ startPostgresProcess time logger CompletePostgresPlan {..} = do
         <$> startProcess "postgres" completePostgresPlanProcessConfig
 
   -- Start postgres and stop if an exception occurs
-  bracketOnError startAction stopPostgresProcess $
+  bracketOnError startAction (stopPostgresProcess False) $
     \result@PostgresProcess {..} -> do
       logger WaitForDB
       -- We assume that 'template1' exist and make connection
@@ -305,14 +307,14 @@ executeInitDb config = do
   (res, stdOut, stdErr) <- executeProcessAndTee "initdb" config
   throwIfNotSuccess (InitDbFailed stdOut stdErr) res
 
-data CopyDirectoryCommand = CopyDirectoryCommand
+data CompleteCopyDirectoryCommand = CompleteCopyDirectoryCommand
   { copyDirectoryCommandSrc :: FilePath
   , copyDirectoryCommandDst :: FilePath
   , copyDirectoryCommandCow :: Bool
   } deriving (Show, Eq, Ord)
 
-instance Pretty CopyDirectoryCommand where
-  pretty CopyDirectoryCommand {..}
+instance Pretty CompleteCopyDirectoryCommand where
+  pretty CompleteCopyDirectoryCommand {..}
     =  text "copyDirectoryCommandSrc:"
     <> softline
     <> indent 2 (text copyDirectoryCommandSrc)
@@ -324,8 +326,8 @@ instance Pretty CopyDirectoryCommand where
     <> text "copyDirectoryCommandCow:"
     <+> (pretty copyDirectoryCommandCow)
 
-executeCopyDirectoryCommand :: CopyDirectoryCommand -> IO ()
-executeCopyDirectoryCommand CopyDirectoryCommand {..} = do
+executeCopyDirectoryCommand :: CompleteCopyDirectoryCommand -> IO ()
+executeCopyDirectoryCommand CompleteCopyDirectoryCommand {..} = do
   let
 #ifdef darwin_HOST_OS
     cpFlags = if copyDirectoryCommandCow then "cp -Rc " else "cp -R "
@@ -347,7 +349,7 @@ executeCopyDirectoryCommand CopyDirectoryCommand {..} = do
 data CompletePlan = CompletePlan
   { completePlanLogger            :: Logger
   , completePlanInitDb            :: Maybe CompleteProcessConfig
-  , completePlanCopy              :: Maybe CopyDirectoryCommand
+  , completePlanCopy              :: Maybe CompleteCopyDirectoryCommand
   , completePlanCreateDb          :: Maybe CompleteProcessConfig
   , completePlanPostgres          :: CompletePostgresPlan
   , completePlanConfig            :: String
@@ -361,6 +363,10 @@ instance Pretty CompletePlan where
     <>  softline
     <>  indent 2 (pretty completePlanInitDb)
     <>  hardline
+    <>  text "completePlanCopy:"
+    <>  softline
+    <>  indent 2 (pretty completePlanCopy)
+    <>  hardline
     <>  text "completePlanCreateDb:"
     <>  softline
     <>  indent 2 (pretty completePlanCreateDb)
@@ -412,11 +418,11 @@ startPlan plan@CompletePlan {..} = do
   let startAction = startPostgresProcess
         completePlanConnectionTimeout completePlanLogger completePlanPostgres
 
-  bracketOnError startAction stopPostgresProcess $ \result -> do
+  bracketOnError startAction (stopPostgresProcess False) $ \result -> do
     for_ completePlanCreateDb executeCreateDb
 
     pure result
 
 -- | Stop the @postgres@ process. See 'stopPostgresProcess' for more details.
 stopPlan :: PostgresProcess -> IO ExitCode
-stopPlan = stopPostgresProcess
+stopPlan = stopPostgresProcess False
